use std::sync::Arc;

use anyhow::Result;

use ui::prelude::*;
use vulkan_rs::prelude::*;

#[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
pub enum BehindLeader {
    Time(f64),
    Laps(i32),
    DNF,
    DSQ,
    PITS,
}

pub struct LeaderBoardEntry {
    id: i32,

    name: String,
    place: u8,
    behind: BehindLeader,
    time_behind_next: f64,
    best_lap: f64,

    snippet: Arc<GuiSnippet>,

    name_label: Arc<Icon>,
    place_label: Arc<Icon>,
    time_label: Arc<Icon>,

    place_updated: bool,
}

impl LeaderBoardEntry {
    const ENTRY: &str = include_str!("leaderboard_entry.xml");

    pub fn create_snippet(gui_handler: &Arc<GuiHandler>) -> Result<Arc<GuiSnippet>> {
        GuiSnippet::from_str(gui_handler, Self::ENTRY)
    }

    pub fn empty(gui_handler: &Arc<GuiHandler>) -> Result<Self> {
        Self::new(
            gui_handler,
            -1,
            "".to_string(),
            0,
            BehindLeader::Laps(0),
            0.0,
            0.0,
        )
    }

    pub fn new(
        gui_handler: &Arc<GuiHandler>,
        id: i32,
        name: String,
        place: u8,
        behind: BehindLeader,
        time_behind_next: f64,
        best_lap: f64,
    ) -> Result<Self> {
        let snippet = Self::create_snippet(gui_handler)?;

        let name_label: Arc<Icon> = snippet.element("name")?;
        let place_label: Arc<Icon> = snippet.element("place")?;
        let time_label: Arc<Icon> = snippet.element("time")?;

        name_label.set_text(&name)?;
        place_label.set_text(place)?;
        time_label.set_text("---")?;

        Ok(Self {
            id,

            name,
            place,
            behind,
            time_behind_next,
            best_lap,

            snippet,

            name_label,
            place_label,
            time_label,

            place_updated: true,
        })
    }

    pub fn change_id(&mut self, id: i32) {
        self.id = id;
    }

    pub fn id(&self) -> i32 {
        self.id
    }

    pub fn place(&self) -> u8 {
        self.place
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn change_name(&mut self, name: String) -> Result<()> {
        self.name = name;

        self.name_label.set_text(Self::check_string(&self.name))
    }

    fn check_string(s: &str) -> String {
        std::str::from_utf8(
            &s.as_bytes()
                .iter()
                .filter(|&&b| {
                    (b > 96 && b < 123) // small letters
                        || (b > 64 && b < 91) // big letters
                        || (b > 47 && b < 58) // numbers
                        || b == 45 // dash
                        || b == 95 // underscore
                        || b == 32 // whitespace
                })
                .map(|&b| b)
                .collect::<Vec<u8>>(),
        )
        .unwrap()
        .to_string()
    }

    pub fn snippet(&self) -> Arc<GuiSnippet> {
        self.snippet.clone()
    }

    pub fn change_background_color(
        &self,
        (place_image, name_image, time_image): &(Arc<Image>, Arc<Image>, Arc<Image>),
    ) -> Result<()> {
        self.time_label.set_icon(&time_image)?;
        self.name_label.set_icon(&name_image)?;
        self.place_label.set_icon(&place_image)?;

        Ok(())
    }

    pub fn update_place(&mut self, place: u8) -> Result<()> {
        if self.place != place {
            self.place_updated = true;

            self.place = place;
            self.place_label.set_text(self.place)?;
        }

        Ok(())
    }

    pub fn reset_time(&mut self) -> Result<()> {
        self.behind = BehindLeader::Time(f64::MIN);
        self.best_lap = f64::MIN;
        self.time_behind_next = f64::MIN;

        self.time_label.set_text("---")
    }

    fn split_minute(time: f64) -> (f64, f64) {
        let full_minutes = (time / 60.0).floor();
        let remainder = time - (full_minutes * 60.0);

        (full_minutes, remainder)
    }

    fn calculate_behind_leader(behind: BehindLeader) -> String {
        match behind {
            BehindLeader::Time(time_behind) => {
                // check if we are leader
                if time_behind <= 0.0 {
                    "---".to_string()
                } else {
                    if time_behind > 60.0 {
                        let (full_minutes, remainder) = Self::split_minute(time_behind);

                        format!("+{:.0}:{:.0}", full_minutes, remainder)
                    } else {
                        format!("+{:.3}", time_behind)
                    }
                }
            }
            BehindLeader::Laps(laps_behind) => {
                format!("+{}", laps_behind)
            }
            BehindLeader::DNF => "DNF".to_string(),
            BehindLeader::DSQ => "DSQ".to_string(),
            BehindLeader::PITS => "PIT".to_string(),
        }
    }

    pub fn force_display_behind_leader(&self) -> Result<()> {
        self.time_label
            .set_text(Self::calculate_behind_leader(self.behind))
    }

    pub fn update_time_behind_leader(&mut self, behind: BehindLeader) -> Result<()> {
        if self.behind != behind {
            self.behind = behind;

            self.force_display_behind_leader()?;
        }

        Ok(())
    }

    fn calculuate_best_lap(best_lap: f64) -> String {
        if best_lap <= 0.0 {
            "---".to_string()
        } else {
            if best_lap > 60.0 {
                let (full_minutes, remainder) = Self::split_minute(best_lap);

                format!("{:.0}:{:.3}", full_minutes, remainder)
            } else {
                format!("{:.3}", best_lap)
            }
        }
    }

    pub fn force_display_best_lap(&mut self) -> Result<()> {
        self.time_label
            .set_text(Self::calculuate_best_lap(self.best_lap))
    }

    pub fn update_best_lap(&mut self, time: f64) -> Result<()> {
        if self.best_lap != time {
            self.best_lap = time;

            self.force_display_best_lap()?;
        }

        Ok(())
    }

    fn calculate_behind_next(behind_next: f64) -> String {
        if behind_next <= 0.0 {
            "---".to_string()
        } else if behind_next > 60.0 {
            let (full_minutes, remainder) = Self::split_minute(behind_next);

            format!("+{:.0}:{:.0}", full_minutes, remainder)
        } else {
            format!("+{:.3}", behind_next)
        }
    }

    pub fn force_display_behind_next(&mut self) -> Result<()> {
        self.time_label
            .set_text(Self::calculate_behind_next(self.time_behind_next))
    }

    pub fn update_time_behind_next(&mut self, time: f64) -> Result<()> {
        if self.time_behind_next != time {
            self.time_behind_next = time;

            self.force_display_behind_next()?;
        }

        Ok(())
    }

    pub fn needs_resorting(&self) -> bool {
        self.place_updated
    }

    pub fn resorting_finished(&mut self) {
        self.place_updated = false;
    }
}

#[cfg(test)]
mod test {
    use super::{BehindLeader, LeaderBoardEntry};

    #[test]
    fn test_string_replacement() {
        let s_orig = "blaü";
        let s = super::LeaderBoardEntry::check_string(s_orig);
        println!("{} {:?}", s_orig, s);

        for b in s.as_bytes() {
            println!("{}", b)
        }

        println!();

        let s2_orig = "az AZ 09 # - _";
        let s2 = super::LeaderBoardEntry::check_string(s2_orig);
        println!("{} {:?}", s2_orig, s2);

        for b in s2.as_bytes() {
            println!("{}", *b);
        }
    }

    #[test]
    #[allow(unused)]
    fn entry_display_tests() {
        let behind_leader = LeaderBoardEntry::calculate_behind_leader(BehindLeader::Time(85.42));
        let behind_next = LeaderBoardEntry::calculate_behind_next(150.213423);
        let best_lap = LeaderBoardEntry::calculuate_best_lap(97.23436);
    }
}
